import 'dart:math';

import 'package:test/test.dart';

import 'arbitrary.dart';
import 'utils.dart';

void glados<T>(
  String name,
  Tester<T> tester, {
  Random random,
  Arbitrary<T> arbitrary,
  int initialExploreSize = 10,
  int exploreCount = 100,
  double exploreSpeed = 1,
}) {
  assert(name != null);
  assert(tester != null);
  random ??= Random(42);
  arbitrary ??= gladosArbitraries.firstWhere(
    (arbitrary) => arbitrary is Arbitrary<T>,
    orElse: () => throw 'glados: No arbitrary found.', // TODO: Better message.
  );
  assert(initialExploreSize != null);
  assert(initialExploreSize > 0);
  assert(exploreCount != null);
  assert(exploreCount > 0);
  assert(exploreSpeed != null);
  assert(exploreSpeed >= 0);

  final stats = Statistics();

  T explore() {
    var tries = 0;
    var size = initialExploreSize.toDouble();

    while (tries < exploreCount) {
      stats.exploreCounter++;
      final input = arbitrary.generate(random, size.ceil());
      if (!succeeds(tester, input)) {
        return input;
      }

      tries++;
      size += exploreSpeed;
    }
    return null;
  }

  T shrink(T initialErrorInducingInput) {
    T input = initialErrorInducingInput;
    bool shrunkSomething = true;

    while (shrunkSomething) {
      shrunkSomething = false;
      for (final shrunkInput in arbitrary.shrink(input)) {
        stats.shrinkCounter++;
        if (!succeeds(tester, shrunkInput)) {
          input = shrunkInput;
          shrunkSomething = true;
          break;
        }
      }
    }
    return input;
  }

  test('$name (testing $exploreCount inputs 🍰)', () {
    final errorInducingInput = explore();

    if (errorInducingInput == null) {
      return; // Let the glados test succeed.
    }

    final shrunkInput = shrink(errorInducingInput);
    print('Tested ${stats.exploreCounter} inputs, shrunk '
        '${stats.shrinkCounter} times.\nFailing for input: $shrunkInput');
    tester(shrunkInput); // This should fail the test again.

    assert(false); // TODO
  });
}
